/**
 * audiocard.c - soundcard handling routines for the RLC-Meter project
 *
 * Copyright (C) 2004 - 2005 Andreas Czechanowski <a.cz @ gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <limits.h>
#include <unistd.h>
#include "audiocard.h"

#if 0
char *dsp_errstr[] = {
  "Device not accessable",
  "Unspecified error",
  "Duplex mode not supported",
  "Triggering not supported",
  "Date formar not supported",
  "Stereo mode not supported",
  "Fragment size not supported",
  "Sampling rate not supported"
};
#endif

const char *MixerDevLabels[] = SOUND_DEVICE_LABELS;
const char *MixerDevNames[] = SOUND_DEVICE_NAMES;

#define STRCOPY(dest,src) { if ((dest) != NULL) { strncpy(dest, src, AUDIO_ERRLEN); } }

/**
 * initializes the DSP access structure.
 *
 * @param D      DSP parameters, structure will be filled by function
 * @param device device name identifying the codec, usually "/dev/dsp0"
 * 
 * @return    0 if everything is Ok, negative value on error.
 */
int
init_dspdata(DspParams *D, char *device)
{
   strncpy(D->device, device, PATH_MAX);
   D->fd = -1;
   D->chunk_size = 0;

   return 0;
}

/**
 * opens the DSP(codec)-device and sets the parameters to the values
 * given in D. The device is left open, and the file descriptor is set
 * in the structure for input/output operations.
 *
 * @param D      parameters to set or change on the DSP device
 * @param errstr if not NULL, error string will be copied here upon return.
 *
 * @return   file descriptor if everything is Ok, negative value on error.
 */
int
open_dsp(DspParams *D, char *errstr)
{
   int dsp_fd;
   char errbuf[AUDIO_ERRLEN + 1];
   int caps;
   int enable_bits;
   int fragsize;

   dsp_fd = open(D->device, O_RDWR);
   if (dsp_fd < 0)
   {
      /* cannot open device - nonexistant device or wrong permissions ? */
      sprintf(errbuf, "open_dsp: Error opening %s: %s\n",
              D->device, strerror(errno));
      STRCOPY(errstr, errbuf);
      return -1;
   }

   if (D->duplex > 0)
   {
      /*
       * switch to duplex mode before even checking if it is available.
       * This seems weird, but is necessary to ensure that GETCAPS
       * does not deny the duplex capability of the device.
       */
      if (ioctl(dsp_fd, SNDCTL_DSP_SETDUPLEX, 0))
      {
         sprintf(errbuf, "open_dsp: cannot activate full-duplex mode\n");
         STRCOPY(errstr, errbuf);
         return -1;
      }
   }

   /* get bitmask with capabilities of the DSP device (should always work) */
   if (ioctl(dsp_fd, SNDCTL_DSP_GETCAPS, &caps))
   {
      sprintf(errbuf, "open_dsp: cannot get capabilities from soundcard\n");
      STRCOPY(errstr, errbuf);
      return -1;
   }

   /*
    * aviod starting the device by a read or write for now
    * before setting parameters
    */
   enable_bits = ~(PCM_ENABLE_OUTPUT|PCM_ENABLE_INPUT);
   if (ioctl(dsp_fd, SNDCTL_DSP_SETTRIGGER, &enable_bits))
   {
      sprintf(errbuf, "open_dsp: cannot disable trigger\n");
      STRCOPY(errstr, errbuf);
      return -1;
   }

   /* set the desired data format for digitized audio */
   if (ioctl(dsp_fd, SNDCTL_DSP_SETFMT, &(D->data_format)))
   {
      sprintf(errbuf, "open_dsp: cannot set data format (%08x)\n",
              D->data_format);
      STRCOPY(errstr, errbuf);
      return -1;
   }

   /* set mono or stereo mode as desired */
   if (ioctl(dsp_fd, SNDCTL_DSP_STEREO, &(D->stereo)))
   {
      sprintf(errbuf, "open_dsp: cannot set stereo mode to %d\n",
              D->stereo);
      STRCOPY(errstr, errbuf);
      return -1;
   }

   /* select the desired sample rate */
   if (ioctl(dsp_fd, SNDCTL_DSP_SPEED, &(D->sample_rate)))
   {
      sprintf(errbuf, "open_dsp: sample rate (%d Hz) not supported\n",
                      D->sample_rate);
      STRCOPY(errstr, errbuf);
      return -1;
   }

   if (D->chunk_size > 64)
   {
      /* set the size (in bytes) of the chunks for sending/receiving data */
      for (fragsize = 6; (1 << fragsize) < D->chunk_size; )
      {
         fragsize++;
      }
      /*
       * the upper bits set the number of chunks to allocate,
       * 0x7fff means as much as possible
       */
      fragsize += 0x7fff0000;
      if (ioctl(dsp_fd, SNDCTL_DSP_SETFRAGMENT, &fragsize))
      {
         sprintf(errbuf, "open_dsp: cannot set fragment size -- "
                 "using default\n");
         STRCOPY(errstr, errbuf);
         /* continue, do not return */
      }
   }

   /* get the actual fragment size used by the soundcard */
   if (ioctl(dsp_fd, SNDCTL_DSP_GETBLKSIZE, &(D->chunk_size)))
   {
      sprintf(errbuf, "open_dsp: cannot determine fragment size\n");
      STRCOPY(errstr, errbuf);
      return -1;
   }

   /* enable starting the DSP device by reading from it (mandatory) */
   enable_bits = PCM_ENABLE_INPUT;
   if (ioctl(dsp_fd, SNDCTL_DSP_SETTRIGGER, &enable_bits))
   {
      sprintf(errbuf, "open_dsp: cannot set trigger on input of DSP\n");
      STRCOPY(errstr, errbuf);
      return -1;
   }

   /* enable starting the DSP device by reading from it (mandatory) */
   enable_bits = PCM_ENABLE_INPUT | PCM_ENABLE_OUTPUT;
   if (ioctl(dsp_fd, SNDCTL_DSP_SETTRIGGER, &enable_bits))
   {
      /*
       * ALSA seems to have a problem with it (returning error),
       * but it does the right thing.
       */
#if 0
      sprintf(errbuf, "open_dsp: cannot set trigger on input/output of DSP\n");
      STRCOPY(errstr, errbuf);
      return -1;
#endif
   }

   D->fd = dsp_fd;

   return dsp_fd;
}

/**
 * initializes dev_labels and copies the device-name into M->device.
 *
 * @param M      mixer parameters, structure will be filled by function
 * @param device device name identifying the mixer, usually "/dev/mixer0"
 * 
 * @return    0 if everything is Ok, negative value on error.
 */
int
init_mixerdata(MixerParams *M, char *device)
{
   M->nrdevices = SOUND_MIXER_NRDEVICES;
   strncpy(M->device, device, PATH_MAX);
   M->devmask = 0;
   M->recmask = 0;
   M->recsrc = 0;

   return 0;
}

/**
 * reads the capabilities and the current settings of the mixer device,
 *
 * @param M      mixer parameters, structure will be filled by function
 * @param errstr if not NULL, error string will be copied here upon return.
 *
 * @return   0 if everything is Ok, negative value on error.
 */
int
get_mixer(MixerParams *M, char *errstr)
{
   int mixer_fd;
   int mask;
   char errbuf[AUDIO_ERRLEN + 1];
   int i;

   mixer_fd = open(M->device, O_RDWR);
   if (mixer_fd < 0)
   {
      /* cannot open device - nonexistant device or wrong permissions ? */
      sprintf(errbuf, "Error opening Mixer %s: %s\n",
              M->device, strerror(errno));
      STRCOPY(errstr, errbuf);
      return -1;
   }

   if (ioctl(mixer_fd, SOUND_MIXER_READ_DEVMASK, &(M->devmask)) < 0)
   {
      sprintf(errbuf, "Cannot get device mask of %s\n", M->device);
      STRCOPY(errstr, errbuf);
      return -2;
   }

   if (ioctl(mixer_fd, SOUND_MIXER_READ_RECMASK, &(M->recmask)) < 0)
   {
      sprintf(errbuf, "Cannot get recording mask of %s\n", M->device);
      STRCOPY(errstr, errbuf);
      return -2;
   }

   if (ioctl(mixer_fd, SOUND_MIXER_READ_RECSRC, &(M->recsrc)) < 0)
   {
      sprintf(errbuf, "Cannot get recording source of %s\n", M->device);
      STRCOPY(errstr, errbuf);
      return -2;
   }

   for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
   {
      mask = (1 << i);
      if (M->devmask & mask)
      {
         ioctl(mixer_fd, MIXER_READ(i), &(M->param[i]));
      }
      else
      {
         M->param[i] = -1;
      }
   }

   close(mixer_fd);

   STRCOPY(errstr, "");

   return 0;
}

/**
 * sets the settings of the mixer device to the values given in M,
 *
 * @param M      mixer parameters, structure must contain valid data
 * @param errstr if not NULL, error string will be copied here upon return.
 *
 * @return   0 if everything is Ok, negative value on error.
 */
int
set_mixer(MixerParams *M, char *errstr)
{
   int mixer_fd;
   int mask;
   char errbuf[AUDIO_ERRLEN + 1];
   int i;

   mixer_fd = open(M->device, O_RDWR);
   if (mixer_fd < 0)
   {
      /* cannot open device - nonexistant device or wrong permissions ? */
      sprintf(errbuf, "Error opening Mixer %s: %s\n",
              M->device, strerror(errno));
      STRCOPY(errstr, errbuf);
      return -1;
   }

   if (ioctl(mixer_fd, SOUND_MIXER_READ_DEVMASK, &(M->devmask)) < 0)
   {
      sprintf(errbuf, "Cannot get device mask of %s\n", M->device);
      STRCOPY(errstr, errbuf);
      return -2;
   }

   for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
   {
      mask = (1 << i);
      if ((M->devmask & mask) && (M->param[i] >= 0))
      {
         ioctl(mixer_fd, MIXER_WRITE(i), &(M->param[i]));
      }
   }
   ioctl(mixer_fd, SOUND_MIXER_WRITE_RECSRC, &(M->recsrc));

   close(mixer_fd);

   return 0;
}

/**
 * searches for a mixer control given by a name.
 *
 * @param name  name of the control to search for
 *
 * @return  index of control or -1 if name was not found.
 */
int
mixer_ctrl_by_name(char *name)
{
   int i;

   for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
   {
      if (0 == strcmp(name, MixerDevNames[i]))
         return i;
   }

   return -1;
}
