/*
 * audiocard.h - header-file for the soundcard control library
 * which is part of the RLCmeter project.
 *
 * Copyright (C) 2004 - 2005 Andreas Czechanowski <a.cz @ gmx.de> 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */

#ifndef __AUDIOCARD_H__
#define __AUDIOCARD_H__

#include <limits.h>
#include <sys/soundcard.h>

#define DSP_DFLT_DEV    "/dev/dsp"
#define MIXER_DFLT_DEV  "/dev/mixer"

#define AUDIO_ERRLEN  255

extern const char *MixerDevLabels[];
extern const char *MixerDevNames[];

/**
 * mixer parameters that may be important for the use of the soundcard
 * as full-duplex input/output device
 */
typedef struct
{
   char device[PATH_MAX];
   /**
    * The name of the parameters, assigned from SOUND_DEVICE_LABELS.
    */
   char *dev_labels[SOUND_MIXER_NRDEVICES];
   /**
    * should be set to SOUND_MIXER_NRDEVICES (number of mixer devices)
    */
   int nrdevices;
   /**
    * parameters which can be adjusted in a given range.
    */
   int param[SOUND_MIXER_NRDEVICES];
   /**
    * parameters set or queried via a bitmask
    */
   int devmask;             /**< SOUND_MIXER_DEVMASK */
   int recmask;             /**< SOUND_MIXER_RECMASK */
   int recsrc;              /**< SOUND_MIXER_RECSRC */
} MixerParams;

/**
 * DSP (A/D and D/A converter) parameters
 */
typedef struct
{
   char device[PATH_MAX];
   int fd;                  /**< file descriptor, if opened */
   int sample_rate;         /**< conversion rate, in Hz */
   int stereo;              /**< 1 for stereo, 0 for mono */
   int duplex;              /**< 1 for full, 0 for half duplex */
   int data_format;         /**< audio data format, AFMT_* */
   int chunk_size;          /**< fragment size, in bytes */
} DspParams;

/* error values returned by init_dsp() */
#define EDSP_ACCESS      -1 /**< device /dev/dsp0 could not be opened */
#define EDSP_OTHER       -2 /**< not further specified error */
#define EDSP_DUPLEX      -3 /**< SNDCTL_DSP_SETDUPLEX failed */
#define EDSP_TRIGGER     -4 /**< SNDCTL_DSP_SETTRIGGER failed */
#define EDSP_FORMAT      -5 /**< SNDCTL_DSP_SETFMT (data format) failed */
#define EDSP_STEREO      -6 /**< SNDCTL_DSP_STEREO (# of channels) failed */
#define EDSP_FRAGMENT    -7 /**< SNDCTL_DSP_SETFRAGMENT failed */
#define EDSP_SMPLRATE    -8 /**< SNDCTL_DSP_SPEED (sample rate) failed */

#define EMIX_ACCESS      -1 /**< device /dev/mixer0 could not be opened */
#define EMIX_OTHER       -2 /**< not further specified error */
#define EMIX_AD_SRC      -3 /**< *_WRITE_RECSRC (A/D input select) failed */
#define EMIX_AD_LEVEL    -4 /**< *_WRITE_IGAIN (A/D level adjust) failed */
#define EMIX_DA_LEVEL    -5 /**< *_WRITE_PCM (D/A level adjust) failed */
#define EMIX_LINE_LEVEL  -6 /**< *_WRITE_LINE (Line-IN level adjust) failed */
#define EMIX_MIX_LEVEL   -7 /**< *_WRITE_IMIX (input mixer level) failed */
#define EMIX_MASTER_VOL  -8 /**< *_WRITE_VOLUME (master output adj.) failed */


/**
 * initializes the DSP access structure.
 *
 * @param D      DSP parameters, structure will be filled by function
 * @param device device name identifying the codec, usually "/dev/dsp0"
 *
 * @return    0 if everything is Ok, negative value on error.
 */
extern int init_dspdata(DspParams *D, char *device);

/**
 * opens the DSP(codec)-device and sets the parameters to the values
 * given in D. The device is left open, and the file descriptor is set
 * in the structure for input/output operations.
 *
 * @param D      parameters to set or change on the DSP device
 * @param errstr if not NULL, error string will be copied here upon return.
 *
 * @return   file descriptor if everything is Ok, negative value on error.
 */
extern int open_dsp(DspParams *D, char *errstr);

/**
 * initializes dev_labels and copies the device-name into M->device
 *
 * @param M      mixer parameters, structure will be filled by function
 * @param device device name identifying the mixer, usually "/dev/mixer0"
 *
 * @return    0 if everything is Ok, negative value on error.
 */
extern int init_mixerdata(MixerParams *M, char *device);

/**
 * reads the capabilities and the current settings of the mixer device,
 *
 * @param M      mixer parameters, structure will be filled by function
 * @param errstr if not NULL, error string will be copied here upon return.
 *
 * @return   0 if everything is Ok, negative value on error.
 */
extern int get_mixer(MixerParams *M, char *errstr);

/**
 * sets the settings of the mixer device to the values given in M,
 *
 * @param M      mixer parameters, structure must contain valid data
 * @param errstr if not NULL, error string will be copied here upon return.
 *
 * @return   0 if everything is Ok, negative value on error.
 */
extern int set_mixer(MixerParams *M, char *errstr);

/**
 * searches for a mixer control given by a name.         
 *
 * @param name  name of the control to search for
 * 
 * @return  index of control or -1 if name was not found.
 */
extern int mixer_ctrl_by_name(char *name);

#if 0
extern int init_dsp(char *devname, int samplerate, int stereo, int format, int *fragsamples);
extern int init_mixer(char *devname, int recsource, int rec_volume, int play_volume);
#endif

#endif /* __AUDIOCARD_H__ */

