/**
 * rlcfile.c - config file handling for the RLC-Meter project
 *
 * Copyright (C) 2004 - 2005 Andreas Czechanowski <a.cz @ gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "utils.h"
#include "audiocard.h"
#include "rlcmeter.h"


/**
 * saves the current settings to disk.
 */
int
save_rlc(char *path, char *name, char *file_loc, MixerParams *M, RLCdata *RLC)
{
   FILE *fp;
   int dev;
   int nf, nz;

   fp = NULL;

   /* if file_loc is already set, try opening this file for writing first */
   if ((NULL != file_loc) && (0 != *file_loc))
   {
      fp = fopen(file_loc, "w+");
   }

   /* try to open the file in any other of the given locations */
   if (NULL == fp)
   {
      fp = pfopen(path, name, "w+", file_loc);
   }

   /* if both failed, return error */
   if (NULL == fp)
      return -1;

   fputs("# RLCmeter 1.0\n#\n", fp);

   /* write the MIXER parameters to the file */
   fputs("# Path to the mixer device to use\n", fp);
   fprintf(fp, "MIXER path %s\n", M->device);

   fputs("# Mixer controls, only the supported ones listed.\n"
         "# Values range from 0 to 100 inclusive. Format:\n"
         "# MIXER Channel Left Right\n", fp);
   for (dev = 0; dev < M->nrdevices; dev++)
   {
      if (M->devmask & (1 << dev))
      {
         fprintf(fp, "MIXER %s %d %d\n", MixerDevNames[dev],
              (M->param[dev] & 0xff00) >> 8,
              (M->param[dev] & 0x00ff));
      }
      if (M->recsrc & (1 << dev))
      {
         fprintf(fp, "MIXER RecSrc %s\n", MixerDevNames[dev]);
      }
   }

   fputs("#\n# Sampling rate for the DSP device, must be supported "
            "by soundcard\n", fp);
   fprintf(fp, "SAMPLERATE %d\n", RLC->sample_freq);

   fputs("#\n# Measurement periode in ms.\n"
         "# Must be multiples of 50, min. 200, max. 5000.\n", fp);
   fprintf(fp, "MEAS_TIME %d\n", RLC->meas_time);

   fputs("#\n# Measurement frequencies in Hz.\n"
         "# Update periode must contain an integer number of cycles. Format:\n"
         "# FREQ N_f  freq\n", fp);
   for (nf = 0; nf < NTEST_FREQS; nf++)
   {
      fprintf(fp, "FREQ %d %.0f\n", nf+1, RLC->test_freq[nf]);
   }

   fputs("#\n# Reference impedances in Ohms. Format:\n"
         "# ZREF N_z  Real Imag\n", fp);
   for (nz = 0; nz < NREF_IMPEDS; nz++)
   {
      fprintf(fp, "ZREF %d %.3f\n", nz+1, RLC->zref[nz]);
   }

   fputs("#\n# Calibration values. Format:\n"
         "# CAL Type N_z N_f  Real Imag\n", fp);
   for (nz = 0; nz < NREF_IMPEDS; nz++)
   {
      for (nf = 0; nf < NTEST_FREQS; nf++)
      {
         fprintf(fp, "CAL Short %d %d  %.4e %.4e\n", nz+1, nf+1,
                 RLC->cal_short[CIDX(nf, nz)].r,
                 RLC->cal_short[CIDX(nf, nz)].i);
         fprintf(fp, "CAL Open  %d %d  %.4e %.4e\n", nz+1, nf+1,
                 RLC->cal_open[CIDX(nf, nz)].r,
                 RLC->cal_open[CIDX(nf, nz)].i);
      }
   }

   fputs("#\n# Channel calibration values. Format:\n"
         "# CAL Chnl N_f  Mag Phase\n", fp);
   for (nf = 0; nf < NTEST_FREQS; nf++)
   {
      fprintf(fp, "CAL Chnl %d  %9.6f %9.6f\n", nf+1,
              RLC->cal_chquot[nf].m, RLC->cal_chquot[nf].p);
   }

   fputs("#\n# end.\n", fp);

   fclose(fp);

   return 0;
}

/**
 * loads the configuration file, interprets the data and fills the values
 * into the appropriate variables.
 */
int
load_rlc(char *path, char *name, char *file_loc, MixerParams *M, RLCdata *RLC)
{
   FILE *fp;
#define LINELEN 80
   char line[LINELEN];
   char *st1, *st2, *st3, *st4;

   fp = NULL;

   /* if file_loc is already set, try opening this file for reading first */
   if ((NULL != file_loc) && (0 != *file_loc))
   {
      fp = fopen(file_loc, "r");
   }

   /* try to open the file in any other of the given locations */
   if (NULL == fp)
   {
      fp = pfopen(path, name, "r", file_loc);
   }

   /* if both failed, return error */
   if (NULL == fp)
      return -1;

   /* read the first line, containig version information */
   if (fgets(line, LINELEN, fp) == NULL)
      return -1;

   /* check for correct file header */
   if (strncmp(line, "# RLCmeter 1.0", 14))
      return -1;

   while (fgets(line, LINELEN, fp) != NULL)
   {
      /* ignore comment lines beginning with a hash */
      if ('#' == line[0])
         continue;

      /* cut the newline character off the end of line */
      st1 = line + strlen(line) - 1;
      if (('\n' == *st1) || ('\r' == *st1))
      {
         *st1 = 0;
      }

      /* get the word up to the first space */
      st1 = strtok(line, " ");
      if (NULL == st1)
         continue;

      if (0 == strcmp(st1, "CAL"))
      {
         int flag = 0;
         int nf;
         int nz;
         int nc;
         Complex cval;
         Polar pval;

         st2 = strtok(NULL, " ");
         if (NULL == st2)
            continue;

         if (0 == strcmp(st2, "Short"))
            flag = 1;
         if (0 == strcmp(st2, "Open"))
            flag = 2;
         if (0 == strcmp(st2, "Chnl"))
            flag = 3;

         if ((flag >= 1) && (flag <= 2))
         {
            /* CAL Open and CAL Short */
            st3 = strtok(NULL, " ");
            if (NULL == st3)
               continue;
            nz = atoi(st3) - 1;

            st3 = strtok(NULL, " ");
            if (NULL == st3)
               continue;
            nf = atoi(st3) - 1;

            if ((nf >= 0) && (nf < NTEST_FREQS)
             && (nz >= 0) && (nz < NREF_IMPEDS))
            {
               nc = CIDX(nf, nz);

               st4 = strtok(NULL, " ");
               if (NULL == st4)
                  continue;
               cval.r = atof(st4);

               st4 = strtok(NULL, " ");
               if (NULL == st4)
                  continue;
               cval.i = atof(st4);

               if (1 == flag)
                  RLC->cal_short[nc] = cval;
               if (2 == flag)
                  RLC->cal_open[nc] = cval;
            }
         }
         else
         if (3 == flag)
         {
            /* CAL Chnl */
            st3 = strtok(NULL, " ");
            if (NULL == st3)
               continue;
            nf = atoi(st3) - 1;

            if ((nf >= 0) && (nf < NTEST_FREQS))
            {
               st4 = strtok(NULL, " ");
               if (NULL == st4)
                  continue;
               pval.m = atof(st4);

               st4 = strtok(NULL, " ");
               if (NULL == st4)
                  continue;
               pval.p = atof(st4);

               RLC->cal_chquot[nf] = pval;
            }
         }
      }
      else
      if (0 == strcmp(st1, "MIXER"))
      {
         st2 = strtok(NULL, " ");
         if (NULL == st2)
            continue;

         if (0 == strcmp(st2, "path"))
         {
            st3 = strtok(NULL, " ");
            if (NULL == st3)
               continue;
            strncpy(M->device, st3, PATH_MAX);
         }
         else
         if (0 == strcmp(st2, "RecSrc"))
         {
            int dev;

            st3 = strtok(NULL, " ");
            if (NULL == st3)
               continue;
            dev = mixer_ctrl_by_name(st3);
            if (dev >= 0)
            {
               M->recsrc = (1 << dev);
            }
         }
         else
         {
            int dev;
            int val1;
            int val2;

            dev = mixer_ctrl_by_name(st2);
            if (dev >= 0)
            {
               st3 = strtok(NULL, " ");
               if (NULL == st3)
                  continue;
               val1 = atoi(st3);

               st3 = strtok(NULL, " ");
               if (NULL == st3)
                  continue;
               val2 = atoi(st3);

               M->param[dev] = ((val1 & 0xff) << 8) + (val2 & 0xff);
            }
         }
      }
      else
      if (0 == strcmp(st1, "FREQ"))
      {
         int nf;
         int freq;

         st2 = strtok(NULL, " ");
         if (NULL == st2)
            continue;
         nf = atoi(st2) - 1;

         if ((nf >= 0) && (nf < NTEST_FREQS))
         {
            st3 = strtok(NULL, " ");
            if (NULL == st3)
               continue;
            freq = atoi(st3);
            if ((freq >= 20) && (freq <= 20000))
               RLC->test_freq[nf] = freq;
         }
      }
      else
      if (0 == strcmp(st1, "MEAS_TIME"))
      {
         int mt;

         st2 = strtok(NULL, " ");
         if (NULL == st2)
            continue;
         mt = atoi(st2);
         if ((mt >= 200) && (mt <= 5000) && ((mt % 50) == 0))
            RLC->meas_time = mt;
      }
      else
      if (0 == strcmp(st1, "SAMPLERATE"))
      {
         int sr;

         st2 = strtok(NULL, " ");
         if (NULL == st2)
            continue;
         sr = atoi(st2);
         if ((sr >= 8000) && (sr <= 96000))
            RLC->sample_freq = sr;
      }
      else
      if (0 == strcmp(st1, "ZREF"))
      {
         int nz;
         double zref;

         st2 = strtok(NULL, " ");
         if (NULL == st2)
            continue;
         nz = atoi(st2) - 1;
         if ((nz >= 0) && (nz < NREF_IMPEDS))
         {
            st3 = strtok(NULL, " ");
            if (NULL == st3)
               continue;
            zref = atof(st3);
            RLC->zref[nz] = zref;
         }
       }
   }

   return 0;
} 
