/*
 * RLC-Meter using the soundard of the PC.
 *
 * Copyright (C) 2004 - 2005 Andreas Czechanowski <a.cz @ gmx.de>
 *
 * Thanks to Hrafnkell Eirksson (hkelle @ hi.is) for his
 * rtfilter program giving me a good template for soundcard control.
 *
 * This program uses a full duplex soundcard and a computer
 * as a signal processor for impedance measurement.
 * 
 * 
 * This program may be used and distributed as is stated in the 
 * GNU General Public Licence (GPL) version 2 (June 1991) or later.
 * I make no guarantees that this program will do anything at all, 
 * if it breaks you get to keep both halves.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <curses.h>
#include <sys/soundcard.h>
#include "utils.h"
#include "audiocard.h"
#include "rlcfile.h"
#include "rlcmeter.h"

char *calpath = DFLT_CALPATH;

int debuglvl = 0;
RLCdata     RLC;
MixerParams Mixer;
DspParams   Dsp;

double window_func[96000];

int freq_index = 0; /* use first frequency after startup */
int zref_index = 1; /* use second Z_ref after startup */
int cal_index;      /* combination of the above, to access calibration slots */

attr_t AP_BACK;
attr_t AP_NORMAL;
attr_t AP_HIGH;
attr_t AP_SELECT;
attr_t AP_INPUT;

/**
 * converts cartesian complex to polar complex variable
 */
Polar *
polar(Complex *x)
{
   static Polar p;
   p.m = sqrt((x->r * x->r) + (x->i * x->i));
   p.p = atan2(x->i, x->r);
   return &p;
}

/**
 * converts polar complex to cartesian complex variable
 */
Complex *
complex(Polar *p)
{
   static Complex x;
   x.r = p->m * cos(p->p);
   x.i = p->m * sin(p->p);
   return &x;
}

/**
 * formats a number into a string honoring scientific power.
 * Multipliers are expressed as letters (f, p, n, u, m, k, M, G, T).
 * @param v     value between 1e-15 and 1e+12
 * @param text  space to write the string into
 *
 * @return      '-' if v is negative, '+' if positive, ' ' (space) if 0.0
 */
char
unitstr(double v, char *text)
{
   char magnif[] = "fpnum kMGT";
   int m_index;
   int sign = 1;
   int digit;
   int cdigit;
   int count;

   /* handle special case "v is NaN" quickly */
   if (isnan(v))
   {
      strcpy(text, " NaN? ");
      return ' ';
   }
   /* handle special case v == 0.0 quickly */
   if (v == 0.0)
   {
      strcpy(text, "0.0000");
      return ' ';
   }
   /* avoid taking log of a negative number */
   if (v < 0)
   {
      v = -v;
      sign = -1;
   }

   /* exponent will be -1 for v < 1.0, +1 for v >= 1000 */
   m_index = 0;
   while ((m_index < 4) && (v > 999.99))
   {
      /* scale down by 1000 as often as possible */
      v *= 0.001;
      m_index++;
   }
   while ((m_index > -5) && (v < 1.000))
   {
      /* scale up by 1000 as often as possible */
      v *= 1000.0;
      m_index--;
   }
   if (v >= 100.0)
   {
      count = v * 100.0;
      cdigit = 2;
   }
   else
   if (v >= 10.0)
   {
      count = v * 1000.0;
      cdigit = 1;
   }
   else
   {
      count = v * 10000.0;
      cdigit = 0;
   }
   for (digit = 0; digit < 5; digit++)
   {
      count = count % 100000;
      *text++ = (count / 10000) + '0';
      count = count * 10;
      if (digit == cdigit)
      {
         *text++ = '.';
      }
   }
   *text++ = magnif[m_index + 5];
   *text = '\0';

   return (sign < 0 ? '-' : '+');
}

/**
 * determine the required size for the transmit/receive buffers.
 * The buffer must at least contain the samples for one fragment,
 * but must contain an integer number of periodes of the sinusodial
 * test signal.
 *
 * @param P  structure of RLC data containing sample rate, fragment size,
 *           measurement time and a list of test frequencies.
 *
 * @return  required buffer size for all desired test frequencies,
 *          in samples.
 */
int
calc_bufsize(RLCdata *P)
{
   int max_mult;
   int max_samples;
   int frag_size;
   int f_sample;
   int meas_time;
   double samples;
   double periodes;
   double freq;
   int fn;
   int ml;

   max_samples = 0;
   f_sample  = P->sample_freq;
   frag_size = P->fragment_size / RLC_BPS;
   meas_time = P->meas_time;

   for (fn = 0; fn < NTEST_FREQS; fn++)
   {
      freq = P->test_freq[fn];

      /*
       * check if the frequency gives an integer number of periodes
       * within the measurement time
       */
      periodes = (freq * meas_time / 1000.0) + 0.00005;
      if ((periodes - floor(periodes)) > 0.0001)
      {
         fprintf(stderr, "Warning: frequency %.3f gives no integer number"
                 " of periodes within a measurement time of %d ms\n",
                 freq, meas_time);
      }

      /* preset number of samples to the maximum allowed value */
      samples = (MAX_SMPL_BUFSIZE - frag_size) + 0.00005;
      /*
       * maximum number with which the sample rate may be multiplied
       * without causing a too big sine/cosine buffer size:
       */
      max_mult = (samples * freq) / (double)f_sample;
      /*fprintf(stderr, "max. multiplicator: %d\n", max_mult);*/

      /* determine nr_samples required for an integer number of periodes */
      for (ml = 1; ml < max_mult; ml++)
      {
         samples = ((ml * f_sample) / freq) + 0.00005;
         /*fprintf(stderr, "trying ml = %d: samples = %.6f\n", ml, samples);*/
         if ((samples - floor(samples)) < 0.0001)
         {
            /* samples is an integer number, which is what we want */
            break;
         }
      }
      if (ml >= max_mult)
      {
         fprintf(stderr, "Warning: frequency %.3f cannot be sampled with"
                         " a rate of %d Hz\n", freq, f_sample);
      }

      /* remember the maximum for all test frequencies */
      if ((int)samples > max_samples)
      {
         max_samples = (int)samples;
      }
   }

   return (frag_size + max_samples);
}

#if 1
/**
 * initializes the sinewave table for the test signal generator.
 * The result is a table containing a stereo signal with 16 bit
 * signed integer samples.
 * Left channel will get the sine, right channel the cosine signal,
 * which can be used for synchronous orthogonal modulation/demodulation.
 *
 * @param table       pointer to array of 16 bit signed integer stereo
 *                    output samples, interleaved (L,R,L,R, ...),
 *                    left channel will be sine, right channel cosine
 * @param count       size of the table to fill, in samples
 * @param freq        frequency of the sinewave, in Hz
 * @param samplerate  sampling frequency, in Hz
 *
 * @return            table modulus, the number of samples in the table
 *                    after which the sinewave exactly repeats.
 */
int
init_sinewave(sint16 *table, int count, double freq, int samplerate)
{
   int max_mult;
   double samples;
   double phi;
   double dphi;
   int ml;
   int s, c;
   int i, j, k;

   /*fprintf(stderr, "init_sinewave: count = %d, freq = %.3f, samplerate = %d\n",
           count, freq, samplerate);*/
   /* preset number of samples to the maximum allowed value */
   samples = count + 0.00005;
   /*
    * maximum number with which the sample rate may be multiplied
    * without causing a too big sine/cosine buffer size:
    */
   max_mult = (samples * freq) / (double)samplerate;

   /* determine nr_samples required for an integer number of periodes */
   for (ml = 1; ml < max_mult; ml++)
   {
      samples = ((ml * samplerate) / freq) + 0.00005;
      if ((samples - floor(samples)) < 0.0001)
      {
         /* samples is an integer number, which is what we want */
         break;
      }
   }
   if (ml >= max_mult)
   {
      fprintf(stderr, "Warning: frequency %.3f cannot be sampled with"
                      " a rate of %d Hz\n", freq, samplerate);
   }

   /* phase increment per sample: */
   dphi = 2.0 * M_PI * freq / (double)samplerate;
   phi = 0.0;
   for (i = 0; i < samples; i++)
   {
      j = i+i;
      s = floor((32767.0 * sin(phi)) + 0.5);
      c = floor((32767.0 * cos(phi)) + 0.5);
      /* left output channel gets sine */
      table[j]   = s;
      /* right output channel gets cosine */
      table[j+1] = c;
      /* increment phase */
      phi += dphi;
   }

   /* forward-copy the rest of the samples */
   k = 0;
   for (i = samples; i < count; i++)
   {
      j = i+i;
      table[j]   = table[k++];
      table[j+1] = table[k++];
   }

   return samples;
}
#else
/**
 * initializes the sinewave table for the test signal generator.
 * The result is a table containing a stereo signal wit 16 bit
 * signed integer samples, both channels having identical signals.
 *
 * @param table       pointer to array of 16 bit signed integer stereo
 *                    output samples, interleaved (L,R,L,R, ...)
 * @param count       size of the table to fill, in samples
 * @param freq        frequency of the sinewave, in Hz
 * @param samplerate  sampling frequency, in Hz
 * @param modul       modulus for table, samples repeat after this number
 * @return            offset in table needed for a 90 degrees phaseshift
 *                    used for synchronous orthogonal demodulation.
 */
int
init_sinewave(sint16 *table, int count, double freq, int samplerate, int modul)
{
   int i,j,s;
   double dphi;
   double periode_f;
   int periode_i;

   /* try to fit an integer number of cycles into the modul */ 
   periode_f = (double)modul / ((double)samplerate / freq);
   periode_i = (int)periode_f;

   if (fabs(periode_f - (double)periode_i) > 0.001)
   {
      /*
       * a warning should be issued here, since one modul contains
       * a fractional count of periodes.
       */
      fprintf(stderr, "Warning: one modul of samples contains fractional"
                      "number of periodes!");
   }
   
   /* phase increment per sample: */
   dphi = 2.0 * M_PI * freq / (double)samplerate;
   for (i = 0; i < count; i++)
   {
      j = i+i;
      s = floor((32767.0 * sin(i * dphi)) + 0.5);
      /* both channels get the same output */
      table[j] = s;
      table[j+1] = s;
   }

   /*
    * calculate the offset between sine and cosine, in samples.
    * This is normally one quarter periode, (fs/4) / freq.
    * If the result is fractional, try adding complete periodes
    * until a fit is reached.
    */
   for (i = 1; i <= (modul/4); i += 4)
   {
      periode_f = (double)(i * samplerate) / (4.0 * freq);
      periode_i = (int)periode_f;
      if (fabs(periode_f - (double)periode_i) < 0.001)
      break;
   }
   if (i > (modul/4))
   {
      /*
       * a warning should be issued here, since the
       * sine to cosine offset is a fractional number.
       */
      fprintf(stderr, "Warning: sine-to-cosine offset is fractional!");
   }
   return periode_i;
}
#endif

/**
 * calculates amplitude and phase from recorded samples of 2 channels.
 * @param samples  16 bit signed integer stereo input samples, interleaved
 * @param count    number of input sample points
 * @param table    pointer to the sinewave table for multiplication,
 *                 as generated by init_sinewave()
 * @param sample_idx  nonvolatile counter to perform an integration over
 *                 an integer number of sinewave cycles.
 * @param nsamples number of samples over which to integrate.
 * @param results  [0] contains ch.1 amplitude, [1] contains ch.1 phase,
 *                 [2] contains ch.2 amplitude, [3] contains ch.2 phase.
 *                 amplitude is normalized to sample count, phase is in
 *                 radians.
 * @return  1 if one integration could be completed and the results
 *          could be updated, 0 else.
 */
int
calc_amplitude(sint16 *samples, int count, sint16 *table,
               int *sample_idx, int nsamples,
               Polar *results)
{
   int i,j,k;
   int result;
   static double re1 = 0.0;
   static double im1 = 0.0;
   static double re2 = 0.0;
   static double im2 = 0.0;
   double amp, phase;

   /* initialize the integration variables on the first call */
   if (0 == *sample_idx)
   {
      re1 = im1 = 0.0;
      re2 = im2 = 0.0;
   }

   result = 0;
   for (i = 0; i < count; i++)
   {
      j = i + i;
      k = j + 1;

      re1 += (samples[j] * table[j]) * window_func[*sample_idx];
      im1 += (samples[j] * table[k]) * window_func[*sample_idx];

      re2 += (samples[j+1] * table[j]) * window_func[*sample_idx];
      im2 += (samples[j+1] * table[k]) * window_func[*sample_idx];

      (*sample_idx)++;
      if (*sample_idx >= nsamples)
      {
         /* we have just enough samples for one integration periode */
         amp = sqrt(re1*re1 + im1*im1) * AMP_SCALE / (double)nsamples;
         phase = atan2(im1, re1);
         results[0].m = amp;   /* re1 * AMP_SCALE / nsamples; */
         results[0].p = phase; /* im1 * AMP_SCALE / nsamples; */

         amp = sqrt(re2*re2 + im2*im2) * AMP_SCALE / (double)nsamples;
         phase = atan2(im2, re2);
         results[1].m = amp;
         results[1].p = phase;

         /* initialize sample counter and integrators for next cycle */
         *sample_idx = 0;
         re1 = im1 = 0.0;
         re2 = im2 = 0.0;

         /* set flag, but continue integrating */
         result = 1;
      }
   }

   return result;
}

int
draw_screen(void)
{
   attrset(AP_NORMAL);
   mvaddstr( 1, 16, "RLC-Meter V0.12 by Andreas Czechanowski, DL4SDC");
   mvaddstr( 3, 4, "(C) channel-q : n.nnnnn @ +nnn.nn           ");
   mvaddstr( 4, 4, "(O) open-cal  : (+n.nnnnx + jn.nnnnx)Ohm     ");
   mvaddstr( 5, 4, "(S) short-cal : (+n.nnnnx + jn.nnnnx)Ohm     ");
   mvaddstr( 7, 4, "(Z) R_source  : nnn.nnxOhm");
   mvaddstr( 8, 4, "(F) f_test    : nnn.nnxHz");
   mvaddstr( 8, 42, "Test-Signal: nn.n % - good");
   mvaddstr(10, 4, "Z = R + jX    : (+n.nnnnx + jn.nnnnx)Ohm");
   mvaddstr(10,48, "Q = n.nnnnx");
   mvaddstr(11, 4, "Y = G + jB    : (+n.nnnnx + jn.nnnnx)S");
   mvaddstr(11,48, "D = n.nnnnx");
   mvaddstr(13, 4, "C series      : +n.nnnnxF -- +n.nnnnxOhm");
   mvaddstr(14, 4, "C parallel    : +n.nnnnxF || +n.nnnnxOhm");
   mvaddstr(16, 4, "L series      : +n.nnnnxH -- +n.nnnnxOhm");
   mvaddstr(17, 4, "L parallel    : +n.nnnnxH || +n.nnnnxOhm");
   mvaddstr(13,46, "(K) Trans.Line short : n.nnnn H");
   mvaddstr(14,46, "(L) Trans.Line open  : n.nnnn F");
   mvaddstr(16,46, "Trans.Line Z0  : n.nnnn Ohm");
   mvaddstr(17,46, "Trans.Line Len : n.nnnn m");
   mvaddstr(19, 4, "(W) write calibration data and mixer settings to file");

   /* let ncurses do a screen-update now */
   refresh();

   return 0;
}

int
show_values(RLCdata *P, int what)
{
   char sgn;
   double val;
   char numstr[16];
   int cal_index;

   cal_index = CIDX(freq_index, zref_index);

   attrset(AP_NORMAL);
   switch(what)
   {
      case CAL_VALUES:
         /* display the calibration values */
         mvprintw(3, 20, "%7.5f @ %+8.3f  ",
            P->cal_chquot[freq_index].m,
            180.0 / M_PI * P->cal_chquot[freq_index].p);
         sgn = unitstr(P->cal_open[cal_index].r, numstr);
         mvprintw(4, 21, "%c%s", sgn, numstr);
         sgn = unitstr(P->cal_open[cal_index].i, numstr);
         mvprintw(4, 30, "%c j%s", sgn, numstr);
         sgn = unitstr(P->cal_short[cal_index].r, numstr);
         mvprintw(5, 21, "%c%s", sgn, numstr);
         sgn = unitstr(P->cal_short[cal_index].i, numstr);
         mvprintw(5, 30, "%c j%s", sgn, numstr);
         sgn = unitstr(P->zref[zref_index], numstr);
         mvprintw(7, 20, "%s", numstr);
         sgn = unitstr(P->test_freq[freq_index], numstr);
         mvprintw(8, 20, "%s", numstr);
         break;

      /*
       * show transmision line parameters
       */
      case TRANS_LINE_VALUES:
         /* display L_short and C_open taken by snapshot */
         sgn = unitstr(P->l_trans.r, numstr);
         mvprintw(13, 69, "%s", numstr);
         sgn = unitstr(P->c_trans.r, numstr);
         mvprintw(14, 69, "%s", numstr);
         /* display calculated length and impedance */
         sgn = unitstr(P->Z_trans, numstr);
         mvprintw(16, 63, "%s", numstr);
         sgn = unitstr(P->len_trans, numstr);
         mvprintw(17, 63, "%s", numstr);
         break;

      /*
       * show only some of the values at a time, to avoid timing problems
       * with slow terminals and/or slow computers.
       */
      case MEAS_VALUES_1:
         /* display Z and Y */
         sgn = unitstr(P->zmeas.r, numstr);
         mvprintw(10, 21, "%c%s", sgn, numstr);
         sgn = unitstr(P->zmeas.i, numstr);
         mvprintw(10, 30, "%c j%s", sgn, numstr);
         sgn = unitstr(P->ymeas.r, numstr);
         mvprintw(11, 21, "%c%s", sgn, numstr);
         sgn = unitstr(P->ymeas.i, numstr);
         mvprintw(11, 30, "%c j%s", sgn, numstr);
         break;

      case MEAS_VALUES_2:
         /* display Q, D  and test signal level */
         val = P->vmeas[1].m * 100.0;
         mvprintw(8, 55, "%5.1f%% - %s ", val,
            (val < 50.0) ? "too low " :
            ((val > 100.0) ? "overload" : "good    "));

         val = P->zmeas.i / P->zmeas.r;
         sgn = unitstr(val, numstr);
         mvprintw(10, 52, "%s", numstr);
         val = 1.0 / val;
         sgn = unitstr(val, numstr);
         mvprintw(11, 52, "%s", numstr);
         break;

      case MEAS_VALUES_3:
         /* display serial and parallel capacitor values */
         sgn = unitstr(P->c_ser.r, numstr);
         mvprintw(13, 20, "%c%s", sgn, numstr);
         sgn = unitstr(P->c_ser.i, numstr);
         mvprintw(13, 33, "%c%s", sgn, numstr);
         sgn = unitstr(P->c_par.r, numstr);
         mvprintw(14, 20, "%c%s", sgn, numstr);
         sgn = unitstr(P->c_par.i, numstr);
         mvprintw(14, 33, "%c%s", sgn, numstr);
         break;

      case MEAS_VALUES_4:
         /* display serial and parallel inductor values */
         sgn = unitstr(P->l_ser.r, numstr);
         mvprintw(16, 20, "%c%s", sgn, numstr);
         sgn = unitstr(P->l_ser.i, numstr);
         mvprintw(16, 33, "%c%s", sgn, numstr);
         sgn = unitstr(P->l_par.r, numstr);
         mvprintw(17, 20, "%c%s", sgn, numstr);
         sgn = unitstr(P->l_par.i, numstr);
         mvprintw(17, 33, "%c%s", sgn, numstr);
         break;
   }

   /* let ncurses do a screen-update now */
   refresh();

   return 0;
}

/**
 * sets some default calibration values to ensure operation without
 * floating point exceptions
 */
int
init_rlc(RLCdata *P)
{
   int i;

   /* set the default sample rate */
   P->sample_freq = SAMPLE_RATE;

   /* set the default fragment size in bytes */
   P->fragment_size = BUF_BYTES;

   /* set an integration time of 500 ms (2 measurements per second) */
   P->meas_time = 500;

   for (i = 0; i < NCAL_POINTS; i++)
   {
      /* assume nearly no series resistance (cable etc.) */
      P->cal_short[i].r = 1.0e-6;
      P->cal_short[i].i = 0.000;
      /* assume nearly infinite parallel resistance (cable, sound input) */
      P->cal_open[i].r = 1.0e+9;
      P->cal_open[i].i = 0.000;
   }
   for (i = 0; i < NTEST_FREQS; i++)
   {
      /* assume test frequencies of 100, 1k and 10k Hz */
      P->test_freq[i] = 100.0 * pow(10.0, (double)i);
      /* assume both channels to be absolutely equal (no amp./phase diff.) */
      P->cal_chquot[i].m = 1.000;
      P->cal_chquot[i].p = 0.000;
   }
   for (i = 0; i < NREF_IMPEDS; i++)
   {
      /* assume reference (series) resistors of 100, 1k and 10k Ohms */
      P->zref[i] = 100.0 * pow(10.0, (double)i);
   }

   /* initialize transmission line data to avoid division by 0 */
   P->l_trans.r = 1.00001e-12;
   P->l_trans.i = 0.0;
   P->c_trans.r = 1.00001e-15;
   P->c_trans.i = 0.0;

   return 0;
}

/**
 * calculates internal correction values from the calibration parameters.
 * These are then used for a faster calculation of the resulting Z.
 * Function should be calles whenever the test frequency or reference
 * impedance have been changed.
 */
int
calc_coeffcal(RLCdata *P)
{
   Polar zi_p, zr_p, qz_p;
   Complex qz_c;

   /* convert input and reference impedance to polar format */
   zi_p = *polar(&(P->cal_open[CIDX(freq_index, zref_index)]));
   zr_p.m = P->zref[zref_index];
   zr_p.p = 0.0;

   /* calculate R_ref / Zi_open */
   qz_p.m = zr_p.m / zi_p.m;
   qz_p.p = zr_p.p - zi_p.p;

   qz_c = *complex(&qz_p);
   qz_c.r += 1.0;

   P->cal_zqout = qz_c;

   return 0;
}

/**
 * calculates the open-correction for measurement of all the
 * impedances in parallel to the unknown Z.
 */
int
calc_opencal(RLCdata *P)
{
   Polar   qu_p;
   Complex qu_c;
   Polar   zx_p;
   Complex zx_c;

   /* calculate quotient of voltages, taking channel difference into account */
   qu_p.m = (P->vmeas[1].m / P->vmeas[0].m) * P->cal_chquot[freq_index].m;
   qu_p.p = (P->vmeas[1].p - P->vmeas[0].p) + P->cal_chquot[freq_index].p;

   /* subtract 1 from this qoutient */
   qu_c = *complex(&qu_p);
   qu_c.r -= 1.0;
   qu_p = *polar(&qu_c);

   /* calculate Zi_open = R_ref / quotient */
   zx_p.m = P->zref[zref_index] / qu_p.m;
   zx_p.p = 0.0                 - qu_p.p;

   /* transform back to rectangular complex */
   zx_c = *complex(&zx_p);
   P->cal_open[CIDX(freq_index, zref_index)] = zx_c;

   return 0;
}

/**
 * does the real computation of the unknown Z.
 */
int
calc_rlc(RLCdata *P)
{
   Polar   qu_p;
   Complex qu_c;
   Polar   zx_p, yx_p;
   Complex zx_c, yx_c;
   double omega;
   int cal_index;

   cal_index = CIDX(freq_index, zref_index);

   /* calculate quotient of voltages, taking channel difference into account */
   qu_p.m = (P->vmeas[1].m / P->vmeas[0].m) * P->cal_chquot[freq_index].m;
   qu_p.p = (P->vmeas[1].p - P->vmeas[0].p) + P->cal_chquot[freq_index].p;

   /* subtract (1 + R_ref / R_open) from this qoutient */
   qu_c = *complex(&qu_p);
   qu_c.r -= P->cal_zqout.r;
   qu_c.i -= P->cal_zqout.i;
   qu_p = *polar(&qu_c);

   /* calculate Z = R_ref / quotient */
   zx_p.m = P->zref[zref_index] / qu_p.m;
   zx_p.p = 0.0                 - qu_p.p;

   /* subtract the short circuit resistance R_short from the result */
   zx_c = *complex(&zx_p);
   zx_c.r -= P->cal_short[cal_index].r;
   zx_c.i -= P->cal_short[cal_index].i;
   zx_p = *polar(&zx_c);

   /* calculate Y =: 1 / Z */
   yx_p.m = 1.0 / zx_p.m;
   yx_p.p = 0.0 - zx_p.p;

   yx_c = *complex(&yx_p);

   P->zmeas = zx_c;
   P->ymeas = yx_c;

   /* transform results into equivalent circuits */
   omega = 2.0 * M_PI* P->test_freq[freq_index];

   P->c_ser.r = -1.0 / (omega * zx_c.i);
   P->c_ser.i = zx_c.r;

   P->c_par.r = yx_c.i / omega;
   P->c_par.i = 1.0 / yx_c.r;

   P->l_ser.r = zx_c.i / omega;
   P->l_ser.i = zx_c.r;

   P->l_par.r = -1.0 / (omega * yx_c.i);
   P->l_par.i = 1.0 / yx_c.r;

   return 0;
}

/**
 * calculates impedance and length of a transmission line from the
 * measured short-inductivity and open-capacitance.
 */
int
calc_transline(RLCdata *P)
{
   double omega;
   double c0;
   double zshort;
   double yopen;
   double zprod;
   double zquot;

   omega = 2.0 * M_PI* P->test_freq[freq_index];
   c0 = 2.9979246e+08;  /* free-space light-speed */

   /*
    * Note that l_ser.r = Ls and l_ser.i = Rs,
    * while     c_par.r = Cp and c_par.i = Rp.
    * Since it is possible that at the relatively low test frequencies
    * the series resistance is higher than the reactance, the
    * formulas only use the reactive component of the measured values
    * to avoid huge measurement errors.
    */

   /* build Z_short and Z_open */
   zshort = omega * P->l_trans.r;
   yopen  = omega * P->c_trans.r;

   /* build Z_prod = sqrt(Z_short * Z_open) */
   zprod = sqrt(zshort / yopen);
   /* build Z_quot = sqrt(-Z_short/Z_open) */
   zquot = sqrt(zshort * yopen);

   /*
    * calculate impedance and length of transmission line.
    * the arctan(zquot) is just for mathematical correctness;
    * in reality the length will normally be many orders below the
    * wavelength, so that zqout alone would be accurate enough.
    * Note that the resulting len_trans is the electrical length in
    * free space and must still be multiplied with the shorting factor
    * k_v to get the mechanical cable length.
    */
   P->len_trans = c0 / omega * atan2(zquot, 1.0);
   P->Z_trans   = zprod;
   
   return 0;
}

/* 
 * Implementation of the DirectForm I filter structure.
 * Inputs are:
 * in:    array of length len of the signal to filter
 * out:   array of length len of the filterd signal (returned)
 * len:   length of the signal arrays
 * b:     the numerator coefficient of the filter
 * a:     the denumerator coefficient of the filter
 * flen:  length of the filter (no of denumerator coeffs)
 * delay: the filter delay registers (used for initial values
 *        and returning final values of the delays)
 */
void filt(sint16 *in, sint16 *out, int len, double b[], double a[], int flen, double delay[]) {
	int i, j, k, sampl;
	double w, outd;

	for(i=0;i<len;i++) {
		k = i+i;
		sampl = in[k];
		w = sampl;  /* remove the DC value and convert to floating point */
		for(j=0;j<(flen-1);j++) {
			w -= a[j]*delay[j];
		} 
		outd = b[0]*w;
		for(j=0;j<(flen-1);j++) {
			outd += b[j+1]*delay[j];
		}
		sampl = outd;  /* restore the DC value and convert to integer */
                if (sampl > 32767)
                   sampl = 32767;
                if (sampl < -32768)
                   sampl = -32768;
                out[k] = sampl;
		for(j=(flen-2);j>0;j--) {
			/* shift the values in the delays one down */
			delay[j] = delay[j-1];
		}
		delay[0] = w;
	}
}

int
main(int argc, char *argv[])
{
   int audio_fd;
   fd_set rs, ws;
   int res;
   int ch;
   char *str;
   sint16 *audio_buf_in;
   sint16 *audio_buf_out;
   sint16 *sine_buf;
   int sin_position;
   int table_size;
   int table_modul;
   int sample_idx;
   int accu_samples;
   audio_buf_info ai;
   /* int need_to_write = 0; */
   int size;
   int quit_program;
   int meas_values = 0;
   char *dspname = DSP_DFLT_DEV;
   char *mixname = MIXER_DFLT_DEV;
   char errstr[AUDIO_ERRLEN + 1];
   char calfile[PATH_MAX];
   int i;
#if 0   
   int sel_returns = 0;
#endif

   while ((ch = getopt(argc, argv, "d:m:")) != EOF)
   {
      switch (ch)
      {
         case 'd':
            dspname = optarg;
            break;
         case 'm':
            mixname = optarg;
            break;
         case 'D':
            debuglvl = atoi(optarg);
            break;
         default:
            printf("Unknown option: -%c (ignored)\n", (char)ch);
            break;
      }
   }

   if ((str = getenv("RLCPATH")) != NULL)
   calpath = str;

   /* initialize the RLC-structure with recent defaults */
   init_rlc(&RLC);

   /* initialize the DSP (codec) data structure */
   init_dspdata(&Dsp, dspname);
   /* initialize the mixer data structure */
   init_mixerdata(&Mixer, mixname);

   /* load the settings from the config file */
   res = load_rlc(calpath, RLC_FILE, calfile, &Mixer, &RLC);
   if (0 == res)
   {
      /* set the mixer parameters with the new values from file */
      res = set_mixer(&Mixer, errstr);
      if (res != 0)
      {
         fprintf(stderr, "Warning: %s\n", errstr);
      }
   }
   else
   {
      /* no config file found, get current mixer settings from driver */
      get_mixer(&Mixer, errstr);
      /* save the current settings into the config file */
      save_rlc(calpath, RLC_FILE, calfile, &Mixer, &RLC);
   }

   /* set DSP parameters, copy some from RLC structure */
   Dsp.sample_rate = RLC.sample_freq;
   Dsp.stereo = 1;
   Dsp.duplex = 1;
   Dsp.data_format = AFMT_S16_NE;
   Dsp.chunk_size = RLC.fragment_size;
   /*
    * Initialize the audio hardware to desired sampling frequency,
    * stereo, signed 16 bit quantization with native endianness.
    */
   audio_fd = open_dsp(&Dsp, errstr);
   if (audio_fd < 0)
   {
      printf("error opening %s: %s\n", dspname, errstr);
      exit(-1);
   }
   /* copy fragment size back to RLC structure (may have been modified) */
   RLC.fragment_size = Dsp.chunk_size;

   /* number of samples over which to integrate the complex amplitude */
   accu_samples = (RLC.sample_freq * RLC.meas_time) / 1000;
   /* initialize the windowing function */
   for (i = 0; i < accu_samples; i++)
   {
#if 0
      /* gaussian low-pass, width 2.5 Sigma */
      expo = 2.0 * (i - (accu_samples/2)) / (double)accu_samples ;
      window_func[i] = exp(-2.5*expo*expo);
#endif
#if 0
      /* cosine halfwave */
      expo = M_PI * (i - (accu_samples/2)) / (double)accu_samples ;
      window_func[i] = cos(expo);
#endif
      /* flat, i.e. no windowing */
      window_func[i] = 1.0; /* 1.0 * cos(expo); */
   }


   /*
    * determine the buffer size required for sample storage
    */
   table_size = calc_bufsize(&RLC);
   /*fprintf(stderr, "buffer size required: %d samples\n", table_size);*/

   audio_buf_in = (sint16*)malloc(RLC.fragment_size * sizeof(char));
   audio_buf_out = (sint16*)malloc(RLC.fragment_size * sizeof(char));

   calc_coeffcal(&RLC);

   sine_buf = (sint16*)malloc(2 * table_size * sizeof(sint16));
   table_modul = init_sinewave(sine_buf, table_size,
                 RLC.test_freq[freq_index],
                 RLC.sample_freq);
   sin_position = 0;
   sample_idx = 0;
   cal_index = CIDX(freq_index, zref_index);

   initscr();  /* initialize curses */
   nonl();     /* do not translate LF to CR/LF on output */
   cbreak();   /* disable line buffering, return typed chars immediately */
   noecho();   /* don't echo typed chars */
   keypad(stdscr, TRUE);       /* enable function and cursor keys */

   /* commonly used screen attributes, may be changed here */
   AP_BACK = A_NORMAL;
   AP_NORMAL = A_NORMAL;
   AP_HIGH = A_BOLD;
   AP_SELECT = A_REVERSE;
   AP_INPUT = A_BOLD;

   bkgdset(AP_BACK);
   /* switch off the cursor for continuous measurements */
   curs_set(0);

   draw_screen();
   show_values(&RLC, CAL_VALUES);

#if 0
   /* set non-blocking mode for stdin */
   fcntl(0, F_GETFL, &flags);
   flags |= O_NONBLOCK;
   fcntl(0, F_SETFL, &flags);
#endif

   FD_ZERO(&rs);
   FD_ZERO(&ws);

   quit_program = 0;

#if 0
   /* TEST; literally copy some data chunks directly from input to output */
   mvprintw(21, 4, "-- testing audio read/write --"); refresh();
   read(audio_fd, audio_buf_in, RLC.fragment_size);
   mvprintw(21, 4, "got chunk #1 of data "); refresh();
   write(audio_fd, audio_buf_in, RLC.fragment_size);
   mvprintw(21, 4, "wrote chunk #1 of data "); refresh();
   read(audio_fd, audio_buf_in, RLC.fragment_size);
   mvprintw(21, 4, "got chunk #2 of data "); refresh();
   write(audio_fd, audio_buf_in, RLC.fragment_size);
   mvprintw(21, 4, "wrote chunk #2 of data "); refresh();
#endif

   while(!quit_program) {
      FD_SET(audio_fd, &rs);
      FD_SET(0, &rs);

#if 0
   {
      int enable_bits;
      /* enable starting the DSP device by reading from it (mandatory) */
      enable_bits = PCM_ENABLE_INPUT | PCM_ENABLE_OUTPUT;
      if (ioctl(audio_fd, SNDCTL_DSP_SETTRIGGER, &enable_bits))
      {
	 /*
         fprintf(stderr, "while_loop: cannot set trigger on input of DSP\n");
         exit(1);
         */
      }
   }
#endif

      if ((res = select(audio_fd+1, &rs, NULL, NULL, NULL))<=0)
      { /* wait for data to read */
         printf("res = %d\n",res);
	 exit(1);
      }
#if 0
      mvprintw(21, 4, "# select-returns: %d ", sel_returns++); refresh();
#endif
      ioctl(audio_fd, SNDCTL_DSP_GETISPACE, &ai);
      if (FD_ISSET(audio_fd, &rs) & (ai.bytes >= RLC.fragment_size))
      {
         /* we really have fragment_size bytes or more - read the samples */
         size = read(audio_fd, audio_buf_in, RLC.fragment_size);

         /* feed data to integrator for amplitude and phase calculation */
         res = calc_amplitude(audio_buf_in, (RLC.fragment_size / RLC_BPS),
                sine_buf + sin_position,
                &sample_idx, accu_samples,
                RLC.vmeas);

         if (res > 0)
         {
            /* we have processed one complete frame, do the math now */
            calc_rlc(&RLC);

            /* trigger the display of the calculated data */
            if ((meas_values > MEAS_VALUES_4) || (meas_values < MEAS_VALUES_1))
               meas_values = MEAS_VALUES_1;

            if (debuglvl >= 1)
            {
               /* log amplitude levels and phase relations to stderr */
               fprintf(stderr, "%9.6f  %9.6f    %9.6f  %9.6f\n",
                  RLC.vmeas[0].m, RLC.vmeas[0].p,
                  RLC.vmeas[1].m, RLC.vmeas[1].p);
            }
#if 0
            mvprintw(19, 4, "%8.4f  %8.4f    %8.4f  %8.4f\n",
                RLC.vmeas[0].m, RLC.vmeas[0].p,
                RLC.vmeas[1].m, RLC.vmeas[1].p);
#endif
         }
         if ((meas_values <= MEAS_VALUES_4) && (meas_values >= MEAS_VALUES_1))
         {
            /* display one set of variables, increment flag for next pass. */
            show_values(&RLC, meas_values);
            meas_values++;
         }

         size = write(audio_fd, sine_buf + sin_position, size);
         sin_position += (RLC.fragment_size / RLC_BPS) * 2;
         sin_position %= table_modul * 2;
      }
      if (FD_ISSET(0, &rs))
      {
         /* character from keyboard was received */
         ch = getch();
         switch(ch)
         {
            case 'c':
               /* perform channel quotient calibration */
               RLC.cal_chquot[freq_index].m =
                  RLC.vmeas[0].m / RLC.vmeas[1].m;
               RLC.cal_chquot[freq_index].p =
                  RLC.vmeas[0].p - RLC.vmeas[1].p;
               calc_coeffcal(&RLC);
               show_values(&RLC, CAL_VALUES);
               break;
            case 's':
               /* perform short-circuit calibration */
               RLC.cal_short[cal_index].r = 0.000;
               RLC.cal_short[cal_index].i = 0.000;
               calc_coeffcal(&RLC);
               calc_rlc(&RLC);
               RLC.cal_short[cal_index] = RLC.zmeas;
               calc_coeffcal(&RLC);
               show_values(&RLC, CAL_VALUES);
               break;
            case 'o':
               /* perform open-circuit calibration */
               calc_opencal(&RLC);
/*
               RLC.cal_open[cal_index] = (Complex){ 1.0e+12, 0.000 };
               calc_coeffcal(&RLC);
               calc_rlc(&RLC);
               RLC.cal_open[cal_index] = RLC.zmeas;
*/
               calc_coeffcal(&RLC);
               show_values(&RLC, CAL_VALUES);
               break;
            case 'f':
               /* select test signal frequency */
               freq_index = (freq_index + 1) % NTEST_FREQS;
               cal_index = CIDX(freq_index, zref_index);
               table_modul = init_sinewave(sine_buf, table_size,
                      RLC.test_freq[freq_index],
                      RLC.sample_freq);
               calc_coeffcal(&RLC);
               show_values(&RLC, CAL_VALUES);
               break;
            case 'z':
               /* select reference impedance */
               zref_index = (zref_index + 1) % NREF_IMPEDS;
               cal_index = CIDX(freq_index, zref_index);
               calc_coeffcal(&RLC);
               show_values(&RLC, CAL_VALUES);
               break;
            case 'k':
                /* take snapshot of L for a shorted transmission-line */
                RLC.l_trans = RLC.l_ser;
                calc_transline(&RLC);
                show_values(&RLC, TRANS_LINE_VALUES);
                break;
            case 'l':
                /* take snapshot of C for a ransmission-line with open end */
                RLC.c_trans = RLC.c_par;
                calc_transline(&RLC);
                show_values(&RLC, TRANS_LINE_VALUES);
                break;
            case 'w':
               /* write calibration values and settings to config file */
            {
               time_t tmr;
               struct tm *ltm;

               /*
                * read all mixer parameters again since they may have been
                * changed by an external application
                */
               res = get_mixer(&Mixer, errstr);

               if (res != 0)
               {
                  mvprintw(21, 4, "%s", errstr);
	          break;
	       }

               /* save the current settings into the config file */
               res = save_rlc(calpath, RLC_FILE, calfile, &Mixer, &RLC);

               if (res >= 0)
               {
                  /* notify the user that the config file has been updated */
                  time(&tmr);
                  ltm = localtime(&tmr);
                  mvprintw(21, 4,
                           "Saved configuration to file at %02d:%02d:%02d",
                           ltm->tm_hour, ltm->tm_min, ltm->tm_sec);
               }
            }
               break;
            case 'q':
               /* quit the RLCmeter program */
               quit_program = 1;
               break;
         }
      }
   }

   close (audio_fd);
   endwin();         /* shut down ncurses */

   exit (0);
}
