/*
 * rlcmeter.h - header-file for the RLC-Meter project
 *
 * Copyright (C) 2004 - 2005 Andreas Czechanowski <a.cz @ gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */

#ifndef __RLC_METER_H__
#define __RLC_METER_H__

#define DBG(msg...) fprintf(stderr, msg);

/**
 * default path to the RLCmeter files
 */
#define DFLT_CALPATH  "~/.rlcmeter:/own/etc:."

/**
 * filename of config and calibration file
 */
#define RLC_FILE "rlcmeter.ini"

/**
 * sample  rate used for the RLC meter, should be high enough to
 * allow test frequencies of 10 kHz
 */
#define SAMPLE_RATE 32000

/** total number of test frequencies */
#define NTEST_FREQS 3

/** total number of reference resistors */
#define NREF_IMPEDS 3

/** total number of calibration points */
#define NCAL_POINTS (NTEST_FREQS * NREF_IMPEDS)

/** index to calibration value from frequency and Zref indices */
#define CIDX(f,z) ((f) + (z)*NTEST_FREQS)

/* maximum number of samples for the sine/cosine table */
#define MAX_SMPL_BUFSIZE 65536

/* set a fragment size of 2^13 =: 8192 bytes (not samples) per default */
#define BUF_BITS  14
#define BUF_BYTES (1 << BUF_BITS)

/** bytes per sample, resulting of 2 channels sampled with 16 bits */
#define RLC_BPS  4

/* 2 channels * 2 bytes per sample */
#define BUF_SMPLS (BUF_BYTES / RLC_BPS)

/*
 * size of sinewave lookup table, which should hold one full cycle
 * for the lowest test frequency plus one quarter wave for cosine
 * plus one buffer size for overlapping
 */
#define SIN_TABLE_SMPLS ((((SAMPLE_RATE*5)/4) / 100) + BUF_SMPLS)

/*
 * maximum offset in sinewave table from which the data is sent
 * to the audio device, in samples.
 */
#define SIN_TABLE_MODUL (SAMPLE_RATE / 100)

/*
 * number of samples over which input data is to be integrated.
 * Should be an integer multiple of SIN_TABLE_MODUL
 */
#define ACCU_SAMPLES (SAMPLE_RATE / 2)

/*
 * amplitude scaling factor, =: 2 / (32768 * 32768)
 */
#define AMP_SCALE 1.8626e-9

typedef signed short int sint16;

#if 0
/* no longer used */
enum { CMPLX_RI=0, CMPLX_MP };
#endif

/** flags controlling the display in show_values() */
enum { CAL_VALUES=0, TRANS_LINE_VALUES,
       MEAS_VALUES_1, MEAS_VALUES_2, MEAS_VALUES_3, MEAS_VALUES_4 };

/** complex number in cartesian notation */
typedef struct
{
   double r; /* real part */
   double i; /* imag. part */
} Complex;

/** complex number in polar notation */
typedef struct
{
   double m; /* magnitude */
   double p; /* phase */
} Polar;

typedef struct
{
   int     sample_freq;              /* sampling frequency / Hz */
   int     fragment_size;            /* samples for one chunk to/from DSP */
   int     meas_time;                /* duration of one cycle / ms */
   double  test_freq[NTEST_FREQS];   /* test frequencies / Hz */
   double  zref[NREF_IMPEDS];        /* reference impedances / Ohms */
   Complex cal_short[NCAL_POINTS];   /* short cal impedance / Ohms */
   time_t  cal_short_t[NCAL_POINTS]; /* time of last short cal */
   Complex cal_open[NCAL_POINTS];    /* open cal impedance / Ohms */
   time_t  cal_open_t[NCAL_POINTS];  /* time of last short cal */
   Polar   cal_chquot[NTEST_FREQS];  /* channel signal quotient */
   time_t  cal_chquot_t[NCAL_POINTS];/* time of last short cal */
   Complex cal_zqout;                /* =: 1 + cal_ref / cal_open */
   Complex zmeas;                    /* measured impedance / Ohms */
   Complex ymeas;                    /* measured admittance / Siemens */
   Polar   vmeas[2];                 /* measured relative AC voltages */
   Complex c_ser;                    /* Z as series circuit C + R */
   Complex c_par;                    /* Z as parallel circuit C || R */
   Complex l_ser;                    /* Z as series circuit L + R */
   Complex l_par;                    /* Z as parallel circuit L || R */
   Complex l_trans;                  /* L_short of transmission line */
   Complex c_trans;                  /* C_open of transmission line */
   double  len_trans;                /* length of transmission line */
   double  Z_trans;                  /* Z_0 of transmission line */
   
} RLCdata;

#endif /*__RLC_METER_H__*/
