/*
 * utils.c - utility functions for file handling etc.
 *
 * Copyright (C) 2004 - 2005 Andreas Czechanowski <a.cz @ gmx.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <limits.h>
#include "utils.h"

/**
 * expands path when it begins with "~/" for home directory or "~user/" for
 * home directory of user.
 *
 * @param cp  string with pathname for input/output, must have
 *            enough space to hold the expanded name.
 */
void
path_expand(char *cp)
{
    char *hdp;
    char *np;
    int hl;
    char user[256];
    struct passwd *pwp;

    if (cp[0] != '~')
        return;
    if (cp[1] == '/' || cp[1] == '\0')
    {
        /* home directory of current user */
        if ((hdp = getenv("HOME")) == NULL)
            return;
        hl = strlen(hdp);
        np = cp + 1;
    }
    else
    {
        /* home directory of any other given user */
        np = strchr(cp + 1, '/');
        if (np == NULL)
            np = cp + strlen(cp);
        strncpy(user, cp + 1, np - (cp + 1));
        user[np - (cp + 1)] = '\0';
        pwp = getpwnam(user);
        if (pwp == NULL)
            return;
        hdp = pwp->pw_dir;
        hl = strlen(hdp);
    }
    memmove(cp + hl, np, strlen(np) + 1);
    memcpy(cp, hdp, hl);
}

/**
 * opens file, searches in the given list of paths.
 * Uses path_expand() to replace '~' with the $HOME direcrory
 *
 * @param path   colon-separated list of directories to search in
 * @param name   file to be opened
 * @param mode   may be "r" or "r+" if file should already exist.
 * @param where  if not NULL, the path+name is stored here upon success
 *
 * @return  FILE pointer as in fopen upon success, NULL if file not found
 */
FILE *
pfopen(char *path, char *name, char *mode, char *where)
{
    char fnam[PATH_MAX];
    char *pp, *sp;
    int len;
    FILE *fp;

    pp = path;
    while (*pp) {
        sp = strchr(pp, DIRSEP);
        if (sp != NULL)
            len = sp - pp;
        else
            len = strlen(pp);
        strncpy(fnam, pp, len);
        fnam[len] = '/';
        strncpy(fnam + len + 1, name, PATH_MAX - len - 2);
        path_expand(fnam);

        fp = fopen(fnam, mode);
        if (fp != NULL) {
            if (where != NULL)
                strcpy(where, fnam);
            return fp;
        }

        pp += len;
        if (sp != NULL)
            pp++;
    }
    if (where != NULL)
        *where = '\0';
    return NULL;
}

